//     ****************************************************************************
//	Readme File for Shutdown.exe Win32 shutdown utility
//
//	Shutdown Version 1.002 July 18, 2000
//
//      Shutdown.dpr - Delphi 5 single file Win32 Msg Loop Type Application
//
//	Copyright 2000 Michael Kelly - Small Software Systems LLC
//
//	www.SmallSoftwareSystems.com
//
//	Shutdown is a System Tray utility.  When run it puts
//	an icon in the System Tray(next to the clock.)
//
//	If you right click or left click on the icon with the mouse
//	a menu pops up with choices to Shutdown.. Restart.. Logoff
//	Windows, Minimize All Windows, Restore Minimized Windows,
//	Unload the Shutdown Icon from the Tray, About, and Cancel.
//
//	Double-clicking the left mouse button on the Icon in the Tray
//	will bring up an Ok/Cancel box warning you are about to Shutdown
//	Windows, just the same as if you selected Shutdown Windows from the
//	Popup Menu.  Of course selecting Cancel will abort the shutdown.
//
//	Likewise, selecting Restart or Logoff commands will invoke
//	a warning box giving you the option to Cancel the command.
//
//	Notes:
//
//      = Installation: Unpack the zip file into an empty folder and
//        drag Shutdown.exe to the desktop or another folder to create
//        a shortcut.
//
//	= If you are on Windows NT then you must have Shutdown priveleges
//	  or selecting Shutdown or Restart will abort with an error message.
//
//	= Applications, such as editors, that may have unsaved data have an
//	  opportunity to prompt you to save the data before shutting down,
//	  restarting or loggin off etc..  If the shutdown appears to be hanging
//	  then you may have a hung application that is not responding to the
//	  WM_CLOSE message.  If this happens try killing the hung app from
//	  the task list by pressing Cntrl-Alt-Del=>End Task.
//
//	= Shutdown is provided as a Freeware Demo for Educational purposes
//	  with NO Warranties Express or Implied.  ** Use At Your Own Risk **
//     ****************************************************************************

Program ShutDown;

Uses
  Windows, Messages, ShellApi, SysUtils;

{$R *.RES}


// Since the entire application is implemented in this file, and is quite
// small, I use more globals than necessary just to keep things simple.

Const
  WinClassName = 'ShutDownWindow'; // class name for the main window
  WM_SHELLNOTIFY = WM_USER + 5; // user-defined msg sent by tray icon.
  TrayIcon = 100; // user-defined id for Tray Icon.
  ShutDownCMD = 7; // Commands for WM_COMMAND message from Popup Menu.
  RestartCMD = 6;
  LogOffCMD = 5;
  MinimizeCMD = 4;
  RestoreCMD = 3;
  UnLoadCMD = 2;
  AboutCMD = 1;
  CancelCMD = 0;
  lParamMin = 0; // values for MinimizeRestore which calls EnumWindows..
  lParamClose = 1; // to do the dirty work of minimizing/restoring windows etc.
  lParamRestore = 2;

Var
  WindowClassAtom : ATOM; // RegisterWindowClass yields an atom if successful.
  WindowClassInfo : WNDCLASSEX; // Class info Data structure for main window.
  HWindow : HWND; // Handle for main application window.
  Msg   : TMsg; // Message variable for the message loop.
  IconData : NotifyIconData; // Data Structure for Shell_NotifyIcon.

// This is the Callback fed to EnumWindows so that windows can be
// minimized/restored/closed.. whatever.  Close is not used in this app.
// Only windows without an owner are processed so that all the desktop
// icons don't get minimized to the task bar.

Function EnumCallback(hw : HWND; lp : LPARAM) : BOOL; Stdcall;
Var
  WindowOwner : HWND;

Begin
  Result := BOOL(1);
  If (hw = GetDesktopWindow) Then
    Exit;
  WindowOwner := GetWindow(hw, GW_OWNER); // only process unowned windows

  Case lp Of

    lParamMin : If (WindowOwner = 0) And IsWindowVisible(hw) Then
        Result := ShowWindow(hw, SW_MINIMIZE);

    lParamRestore : If (WindowOwner = 0) And IsWindowVisible(hw) And
      IsIconic(hw) Then
        Result := ShowWindow(hw, SW_RESTORE);

    lParamClose :
      If (WindowOwner = 0) And (hw <> HWindow) And IsWindowVisible(hw) Then
        Result := PostMessage(hw, WM_CLOSE, 0, 0);

  End; // Case
End; // EnumCallback


// This does the call to EnumWindows with the user-defined param
// telling the callback which operation to perform when iterating
// windows on the desktop.  Note that turning off animation and
// locking the desktop while doing so avoids the unpleasant effect
// of traces flying everywhere when minimizing or restoring windows.

Procedure MinimizeRestore(Operation : LPARAM);
Var
  AnimateWasOn, EnumError : Boolean;
  info  : ANIMATIONINFO;

Begin
  EnumError := False;
  info.cbSize := sizeof(info);
  SystemParametersInfo(SPI_GETANIMATION, sizeof(info), @info, 0);
  AnimateWasOn := info.iMinAnimate <> 0;
  info.iMinAnimate := 0;
  SystemParametersInfo(SPI_SETANIMATION, sizeof(info), @info, 0);
  LockWindowUpdate(GetDesktopWindow);
  If EnumWindows(@EnumCallback, Operation) = BOOL(0) Then
  Begin
    EnumError := True;
    LockWindowUpdate(0);
    MessageBox(0, 'Error Minimizing or Restoring Windows!', Nil,
      MB_OK Or MB_ICONWARNING Or MB_SETFOREGROUND);
  End;
  If AnimateWasOn Then
  Begin
    info.iMinAnimate := 1;
    SystemParametersInfo(SPI_SETANIMATION, sizeof(info), @info, 0);
  End;
  If Not EnumError Then
    LockWindowUpdate(0);
End; // MinimizeAll

// if not running on NT, returns True
// else returns the result of trying
// to get the NT Shutdown Priv

Function GetShutDownPriv : Boolean;
Const
  PrivName : PAnsiChar = 'SeShutdownPrivilege';
Var
  hToken : THandle;
  tkp, prev : TTokenPrivileges;
  retlength : DWORD;

Begin
  Result := True;
  If Win32Platform <> VER_PLATFORM_WIN32_NT Then
    Exit; // just return True if on Win9x since there's no security
  Result := False;  // return False if security token cannot be adjusted on NT
  If Not OpenProcessToken(GetCurrentProcess,
    TOKEN_ADJUST_PRIVILEGES Or TOKEN_QUERY,
    hToken) Then
    Exit;
  If Not LookupPrivilegeValue(Nil, PrivName, tkp.Privileges[0].Luid) Then
    Exit;
  prev := tkp;
  tkp.PrivilegeCount := 1;
  tkp.Privileges[0].Attributes := 2;
  AdjustTokenPrivileges(hToken, BOOL(FALSE), tkp, sizeof(prev), prev, retlength);
  If GetLastError <> ERROR_SUCCESS Then
    Exit;
  Result := True; // return True if SeShutdownPrivilege accepted
End; // GetShutDownPriv

// A simple Case statement to invoke ExitWindowsEx with the
// proper param to either shutdown, restart, or log off.
// Note that for shutdown and restart the GetShutDownPriv is called
// in case the utility is running on NT.

Procedure DoWindowsShutdown(Command : WPARAM);
Begin
  Case Command Of

    ShutDownCMD : Begin
        If Not GetShutDownPriv Then
        Begin
          MessageBox(0, 'Could Not get Privelege to Shut Down System : Aborting',
            Nil, MB_OK Or MB_ICONSTOP Or MB_SETFOREGROUND);
          PostQuitMessage(0);
          Exit;
        End;
        If MessageBox(0, 'Continuing will Shutdown Windows!', 'Warning!',
          MB_OKCANCEL Or MB_ICONWARNING Or MB_SETFOREGROUND) = ID_OK Then
          ExitWindowsEx(EWX_SHUTDOWN, 0); // give user chance to cancel Shutdown
      End; // ShutDownCMD

    RestartCMD : Begin
        If Not GetShutDownPriv Then
        Begin
          MessageBox(0, 'Could Not get Privelege to Restart System : Aborting',
            Nil, MB_OK Or MB_ICONSTOP Or MB_SETFOREGROUND);
          PostQuitMessage(0);
          Exit;
        End;
        If MessageBox(0, 'Continuing will Restart Windows!', 'Warning!',
          MB_OKCANCEL Or MB_ICONWARNING Or MB_SETFOREGROUND) = ID_OK Then
          ExitWindowsEx(EWX_REBOOT, 0); // likewise may cancel Restart
      End; // RestartCMD

    LogOffCMD : Begin
        If MessageBox(0, 'Continuing will Log Off Windows!', 'Warning!',
          MB_OKCANCEL Or MB_ICONWARNING Or MB_SETFOREGROUND) = ID_OK Then
          ExitWindowsEx(EWX_LOGOFF, 0); // likewise may cancel Logoff
      End; // LogOffCMD
  End; // Case
End; // DoWindowsShutdown

// Uses the Shell_NotifyIcon Api to put the icon in the System Tray.

Function PutIconInTray : Boolean;
Begin
  IconData.cbSize := sizeof(IconData);
  IconData.Wnd := HWindow;
  IconData.uId := TrayIcon;
  IconData.uFlags := NIF_ICON + NIF_TIP + NIF_MESSAGE;
  IconData.uCallbackMessage := WM_SHELLNOTIFY;
  IconData.hIcon := WindowClassInfo.hIconSm;
  IconData.szTip := 'ShutDown/Logoff or Minimize/Restore Windows';
  Result := Shell_NotifyIcon(NIM_ADD, @IconData);
End; // PutIconInTray

// If About.html is in the working directory, display it and return True
// in response to the About command from the Popup Menu.

Function ShowAboutImage : Boolean;
Var
  FileName : String;

Begin
  Result := False;
  FileName := ExtractFilePath(ParamStr(0)) + 'About.html';
  If Not FileExists(FileName) Then
    Exit;
  Result := True;
  ShellExecute(0, 'open', PChar(FileName), Nil, Nil, SW_SHOW);
End; // ShowAboutImage

// The Win32 message loop WindowProc for the application.
// If the user clicks the Icon a PopUp Menu is created and
// the user selection processed.  A double-click invokes
// the shutdown command.

Function WinProc(hw : HWND; uMsg : UINT; wp : WPARAM; lp : LPARAM) : LRESULT; Stdcall; Export;
Var
  PopHandle : HMENU;
  pnt   : TPoint;

Begin
  Result := 0;
  Case uMsg Of

    WM_CLOSE, WM_DESTROY : Begin
        Shell_NotifyIcon(NIM_DELETE, @IconData);
        PostQuitMessage(0);
      End; // WM_CLOSE : WM_DESTROY

    WM_COMMAND :
      If (lp = 0) And (wp In [ShutDownCMD, RestartCMD, LogOffCMD]) Then
        DoWindowsShutdown(wp)
      Else If (lp = 0) And (wp = AboutCMD) Then
      Begin
        If Not ShowAboutImage Then
          ShellAbout(HWindow, 'Shutdown Utility v 1.002',
            '(c) 2000 Small Software Systems LLC'#13#10'www.SmallSoftwareSystems.com',
            WindowClassInfo.hIcon);
      End
      Else If (lp = 0) And (wp = MinimizeCMD) Then
        MinimizeRestore(lParamMin)
      Else If (lp = 0) And (wp = RestoreCMD) Then
        MinimizeRestore(lParamRestore)
      Else If (lp = 0) And (wp = UnLoadCMD) Then
      Begin
        Shell_NotifyIcon(NIM_DELETE, @IconData);
        PostQuitMessage(0);
      End;
        // WM_COMMAND

    WM_SHELLNOTIFY : Begin
        If wp = TrayIcon Then
        Begin
          If lp = WM_LBUTTONDBLCLK Then
            DoWindowsShutdown(ShutdownCMD)
          Else If (lp = WM_RBUTTONDOWN) Or (lp = WM_LBUTTONDOWN) Then
          Begin
            PopHandle := CreatePopUpMenu;
            If PopHandle = 0 Then
            Begin
              MessageBox(0, 'Unable to Create PopUp Menu : Aborting!',
                Nil, MB_OK Or MB_ICONSTOP Or MB_SETFOREGROUND);
              PostQuitMessage(0);
              Exit;
            End; // If
            AppendMenu(PopHandle, MF_STRING, ShutDownCMD, 'Shut Down Windows');
            AppendMenu(PopHandle, MF_STRING, RestartCMD, 'Restart Windows');
            AppendMenu(PopHandle, MF_STRING, LogOffCMD, 'Log Off Windows');
            AppendMenu(PopHandle, MF_STRING, MinimizeCMD, 'Minimize All Windows');
            AppendMenu(PopHandle, MF_STRING, RestoreCMD, 'Restore Minimized Windows');
            AppendMenu(PopHandle, MF_STRING, UnLoadCMD, 'Unload Program');
            AppendMenu(PopHandle, MF_STRING, AboutCMD, 'About Shutdown');
            AppendMenu(PopHandle, MF_STRING, CancelCMD, 'Cancel');
            Windows.GetCursorPos(pnt);
            TrackPopUpMenu(PopHandle, TPM_RIGHTALIGN, pnt.x, pnt.y, 0, HWindow, Nil);
            DestroyMenu(PopHandle);
          End; // Else If
        End; // if
      End; // WM_SHELLNOTIFY
  Else
    Result := DefWindowProc(hw, uMsg, wp, lp);
  End; // Case
End; // WinProc

// Registers and Creates the main application window.

Function CreateAppWindow : Boolean;
  Function RegisterWindowClass : Boolean;
  Begin
    WindowClassInfo.cbSize := sizeof(WindowClassInfo);
    WindowClassInfo.Style := CS_VREDRAW Or CS_HREDRAW;
    WindowClassInfo.lpfnWndProc := @WinProc;
    WindowClassInfo.cbClsExtra := 0;
    WindowClassInfo.cbWndExtra := 0;
    WindowClassInfo.hInstance := HInstance;
    WindowClassInfo.hIcon := LoadIcon(HInstance, 'MAINICON');
    WindowClassInfo.hCursor := LoadCursor(0, IDC_ARROW);
    WindowClassInfo.hbrBackground := HBrush(COLOR_APPWORKSPACE);
    WindowClassInfo.lpszMenuName := Nil;
    WindowClassInfo.lpszClassName := WinClassName;
    WindowClassInfo.hIconSm := LoadIcon(HInstance, 'SMALICON');
    WindowClassAtom := RegisterClassEx(WindowClassInfo);
    Result := WindowClassAtom <> 0;
  End; // RegisterWindowClass - Nested Function

Begin
  If Not RegisterWindowClass Then
  Begin
    MessageBox(0, 'Unable to Register Main Window : Aborting!',
      Nil, MB_OK Or MB_ICONSTOP);
    Halt;
  End;
  HWindow := CreateWindowEx(WS_EX_OVERLAPPEDWINDOW, WinClassName, 'ShutDown',
    WS_OVERLAPPEDWINDOW, 0, 0, 200, 200, 0, 0, HInstance, Nil);
  If HWindow <> 0 Then
  Begin
    ShowWindow(HWindow, SW_HIDE);
    UpdateWindow(HWindow);
    Result := True;
    Exit;
  End;
  Result := False;
End; // CreateAppWindow

// The application's main entry point.
// The application's main window is created
// then hiden once the icon is successfully
// inserted in the System Tray.  A plain 200x200
// pixel window is used to conserve resources.
// Instead of adorning the window, we rely on
// the About command to display our company's
// web page image, a link to the page, and of
// course copyright information.

Begin
  If Not CreateAppWindow Then
  Begin
    MessageBox(0, 'Unable to Create Main Application Window : Aborting!',
      Nil, MB_OK Or MB_ICONSTOP Or MB_SETFOREGROUND);
    Halt;
  End;
  If Not PutIconInTray Then
  Begin
    MessageBox(0, 'Unable to Put Icon in Task Tray : Aborting!',
      Nil, MB_OK Or MB_ICONSTOP Or MB_SETFOREGROUND);
    Halt;
  End; // if
  While GetMessage(Msg, HWindow, 0, 0) Do  // process messages until Quit
  Begin
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  End;
  Halt(Msg.wParam);
End.

